"""Unit tests for SSE utilities and TreeState builder."""

from unittest.mock import patch

from strawpot_gui.sse import TreeState, _compose_activity, format_sse, sse_retry


class TestFormatSSE:
    def test_basic_format(self):
        result = format_sse(1, {"hello": "world"})
        assert result == 'id: 1\ndata: {"hello":"world"}\n\n'

    def test_retry_format(self):
        result = sse_retry(5000)
        assert result == "retry: 5000\n\n"

    def test_default_retry(self):
        result = sse_retry()
        assert result == "retry: 3000\n\n"


class TestTreeStateSessionJson:
    def test_load_root_agent(self):
        state = TreeState()
        state.load_session_json({
            "agents": {
                "agent_root": {
                    "role": "orchestrator",
                    "runtime": "strawpot-claude-code",
                    "parent": None,
                    "started_at": "T0",
                }
            }
        })
        result = state.to_dict()
        assert len(result["nodes"]) == 1
        assert result["nodes"][0]["agent_id"] == "agent_root"
        assert result["nodes"][0]["status"] == "running"

    def test_load_ignores_existing_agents(self):
        state = TreeState()
        state.load_session_json({
            "agents": {
                "agent_root": {
                    "role": "orchestrator",
                    "runtime": "strawpot-claude-code",
                    "parent": None,
                    "started_at": "T0",
                }
            }
        })
        # Loading again shouldn't overwrite
        state.load_session_json({
            "agents": {
                "agent_root": {
                    "role": "different",
                    "runtime": "different",
                    "parent": None,
                    "started_at": "T1",
                }
            }
        })
        assert state.to_dict()["nodes"][0]["role"] == "orchestrator"

    def test_empty_agents(self):
        state = TreeState()
        state.load_session_json({"agents": {}})
        assert state.to_dict()["nodes"] == []


class TestTreeStateTraceEvents:
    def test_delegate_start_creates_pending(self):
        state = TreeState()
        state.load_session_json({
            "agents": {
                "agent_root": {
                    "role": "orchestrator",
                    "runtime": "cc",
                    "parent": None,
                    "started_at": "T0",
                }
            }
        })
        state.process_event({
            "event": "session_start", "span_id": "s0", "data": {},
        })
        state.process_event({
            "event": "delegate_start", "span_id": "s1",
            "parent_span": "s0",
            "data": {"role": "implementer"},
        })
        result = state.to_dict()
        assert len(result["pending_delegations"]) == 1
        assert result["pending_delegations"][0]["role"] == "implementer"
        assert result["pending_delegations"][0]["requested_by"] == "agent_root"

    def test_agent_spawn_promotes_pending(self):
        state = TreeState()
        state.load_session_json({
            "agents": {
                "agent_root": {
                    "role": "orchestrator",
                    "runtime": "cc",
                    "parent": None,
                    "started_at": "T0",
                }
            }
        })
        state.process_event({
            "event": "session_start", "span_id": "s0", "data": {},
        })
        state.process_event({
            "event": "delegate_start", "span_id": "s1",
            "parent_span": "s0",
            "data": {"role": "implementer"},
        })
        state.process_event({
            "event": "agent_spawn", "span_id": "s1", "ts": "T2",
            "data": {"agent_id": "agent_impl", "runtime": "cc", "pid": 999},
        })
        result = state.to_dict()
        assert len(result["pending_delegations"]) == 0
        agent_ids = [n["agent_id"] for n in result["nodes"]]
        assert "agent_impl" in agent_ids
        impl = next(n for n in result["nodes"] if n["agent_id"] == "agent_impl")
        assert impl["parent"] == "agent_root"
        assert impl["status"] == "running"

    def test_agent_end_marks_completed(self):
        state = TreeState()
        state.process_event({
            "event": "session_start", "span_id": "s0", "data": {},
        })
        state.process_event({
            "event": "delegate_start", "span_id": "s1",
            "parent_span": "s0",
            "data": {"role": "impl"},
        })
        state.process_event({
            "event": "agent_spawn", "span_id": "s1", "ts": "T1",
            "data": {"agent_id": "a1", "runtime": "cc", "pid": 1},
        })
        state.process_event({
            "event": "agent_end", "span_id": "s1",
            "data": {"exit_code": 0, "duration_ms": 5000},
        })
        node = state.nodes["a1"]
        assert node.status == "completed"
        assert node.exit_code == 0
        assert node.duration_ms == 5000

    def test_agent_end_nonzero_marks_failed(self):
        state = TreeState()
        state.process_event({
            "event": "delegate_start", "span_id": "s1",
            "parent_span": None,
            "data": {"role": "impl"},
        })
        state.process_event({
            "event": "agent_spawn", "span_id": "s1", "ts": "T1",
            "data": {"agent_id": "a1", "runtime": "cc", "pid": 1},
        })
        state.process_event({
            "event": "agent_end", "span_id": "s1",
            "data": {"exit_code": 1, "duration_ms": 3000},
        })
        assert state.nodes["a1"].status == "failed"

    def test_delegate_denied_recorded(self):
        state = TreeState()
        state.process_event({
            "event": "delegate_denied", "span_id": "s2",
            "parent_span": "s0",
            "data": {"role": "admin", "reason": "DENY_DEPTH_LIMIT"},
        })
        result = state.to_dict()
        assert len(result["denied_delegations"]) == 1
        assert result["denied_delegations"][0]["role"] == "admin"
        assert result["denied_delegations"][0]["reason"] == "DENY_DEPTH_LIMIT"

    def test_agent_spawn_uses_event_role_when_no_pending(self):
        """Root agent_spawn (no prior delegate_start) reads role from event data."""
        state = TreeState()
        state.process_event({
            "event": "session_start", "span_id": "s0", "data": {},
        })
        state.process_event({
            "event": "agent_spawn", "span_id": "s0", "ts": "T0",
            "data": {
                "agent_id": "agent_root",
                "role": "ai-ceo",
                "runtime": "strawpot-claude-code",
                "pid": 123,
            },
        })
        assert state.nodes["agent_root"].role == "ai-ceo"

    def test_session_end_marks_terminal(self):
        state = TreeState()
        state.process_event({
            "event": "session_end", "span_id": "s0",
            "data": {"duration_ms": 1000},
        })
        assert state.is_terminal

    def test_session_end_completes_root(self):
        state = TreeState()
        state.load_session_json({
            "agents": {
                "agent_root": {
                    "role": "orchestrator",
                    "runtime": "cc",
                    "parent": None,
                    "started_at": "T0",
                }
            }
        })
        state.process_event({
            "event": "session_end", "span_id": "s0",
            "data": {"duration_ms": 5000},
        })
        assert state.nodes["agent_root"].status == "completed"
        assert state.nodes["agent_root"].duration_ms == 5000


class TestTreeStateFullLifecycle:
    def test_full_scenario(self):
        state = TreeState()
        state.load_session_json({
            "agents": {
                "agent_root": {
                    "role": "orchestrator",
                    "runtime": "strawpot-claude-code",
                    "parent": None,
                    "started_at": "T0",
                    "pid": 100,
                }
            }
        })
        events = [
            {"event": "session_start", "span_id": "s0",
             "data": {"run_id": "r1", "role": "orchestrator"}},
            {"event": "delegate_start", "span_id": "s1", "parent_span": "s0",
             "data": {"role": "implementer"}},
            {"event": "agent_spawn", "span_id": "s1", "ts": "T1",
             "data": {"agent_id": "agent_impl", "runtime": "cc", "pid": 200}},
            {"event": "delegate_start", "span_id": "s2", "parent_span": "s1",
             "data": {"role": "reviewer"}},
            {"event": "delegate_denied", "span_id": "s3", "parent_span": "s1",
             "data": {"role": "admin", "reason": "DENY_DEPTH_LIMIT"}},
            {"event": "agent_spawn", "span_id": "s2", "ts": "T2",
             "data": {"agent_id": "agent_rev", "runtime": "cc", "pid": 300}},
            {"event": "agent_end", "span_id": "s2",
             "data": {"exit_code": 0, "duration_ms": 10000}},
            {"event": "delegate_end", "span_id": "s2",
             "data": {"exit_code": 0,
                       "duration_ms": 10000}},
            {"event": "agent_end", "span_id": "s1",
             "data": {"exit_code": 0, "duration_ms": 30000}},
            {"event": "delegate_end", "span_id": "s1", "parent_span": None,
             "data": {"exit_code": 0,
                       "duration_ms": 30000}},
            {"event": "session_end", "span_id": "s0",
             "data": {"duration_ms": 35000}},
        ]
        for ev in events:
            state.process_event(ev)

        result = state.to_dict()

        # 3 nodes: root, implementer, reviewer
        assert len(result["nodes"]) == 3
        ids = {n["agent_id"] for n in result["nodes"]}
        assert ids == {"agent_root", "agent_impl", "agent_rev"}

        # All completed
        for node in result["nodes"]:
            assert node["status"] == "completed"

        # Reviewer's parent is implementer
        rev = next(n for n in result["nodes"] if n["agent_id"] == "agent_rev")
        assert rev["parent"] == "agent_impl"

        # Implementer's parent is root
        impl = next(
            n for n in result["nodes"] if n["agent_id"] == "agent_impl"
        )
        assert impl["parent"] == "agent_root"

        # No pending delegations
        assert len(result["pending_delegations"]) == 0

        # 1 denied delegation
        assert len(result["denied_delegations"]) == 1
        assert result["denied_delegations"][0]["role"] == "admin"

        assert state.is_terminal


class TestTreeStateCancelEvents:
    """Tests for cancel event processing in TreeState."""

    def test_cancel_start_sets_cancelling(self):
        state = TreeState()
        state.load_session_json({
            "agents": {
                "a1": {"role": "orch", "runtime": "cc", "parent": None},
                "a2": {"role": "worker", "runtime": "cc", "parent": "a1"},
            }
        })
        state.process_event({
            "event": "agent_cancel_start",
            "span_id": "s0",
            "data": {"agent_id": "a1", "descendants": ["a2"]},
        })
        assert state.nodes["a1"].status == "cancelling"
        assert state.nodes["a2"].status == "cancelling"

    def test_cancel_complete_sets_cancelled(self):
        state = TreeState()
        state.load_session_json({
            "agents": {
                "a1": {"role": "orch", "runtime": "cc", "parent": None},
                "a2": {"role": "worker", "runtime": "cc", "parent": "a1"},
            }
        })
        state.process_event({
            "event": "agent_cancel_complete",
            "span_id": "s0",
            "data": {"cancelled_agents": ["a2", "a1"]},
        })
        assert state.nodes["a1"].status == "cancelled"
        assert state.nodes["a2"].status == "cancelled"

    def test_cancel_unknown_agents_ignored(self):
        state = TreeState()
        # Should not raise on unknown agent IDs
        state.process_event({
            "event": "agent_cancel_start",
            "span_id": "s0",
            "data": {"agent_id": "unknown", "descendants": ["also_unknown"]},
        })
        state.process_event({
            "event": "agent_cancel_complete",
            "span_id": "s0",
            "data": {"cancelled_agents": ["unknown"]},
        })


class TestTreeStateToolActivity:
    """Tests for tool_start/tool_end activity tracking in TreeState."""

    def _make_state_with_running_agent(self, agent_id="a1"):
        """Helper: create a TreeState with one running agent."""
        state = TreeState()
        state.process_event({
            "event": "delegate_start", "span_id": "s1",
            "parent_span": None, "data": {"role": "impl"},
        })
        state.process_event({
            "event": "agent_spawn", "span_id": "s1", "ts": "T1",
            "data": {"agent_id": agent_id, "runtime": "cc", "pid": 1},
        })
        return state

    def test_tool_start_sets_current_activity_from_summary(self):
        """tool_start with a summary sets current_activity to the summary."""
        state = self._make_state_with_running_agent()
        state.process_event({
            "event": "tool_start", "span_id": "s2",
            "data": {"agent_id": "a1", "tool": "Bash", "summary": "Running tests"},
        })
        assert state.nodes["a1"].current_activity == "Running tests"

    def test_tool_start_falls_back_to_tool_name(self):
        """tool_start without summary uses the tool name as activity."""
        state = self._make_state_with_running_agent()
        state.process_event({
            "event": "tool_start", "span_id": "s2",
            "data": {"agent_id": "a1", "tool": "Read", "summary": ""},
        })
        assert state.nodes["a1"].current_activity == "Read"

    def test_tool_start_with_empty_tool_and_summary_sets_none(self):
        """tool_start with both empty tool and summary sets None."""
        state = self._make_state_with_running_agent()
        state.process_event({
            "event": "tool_start", "span_id": "s2",
            "data": {"agent_id": "a1", "tool": "", "summary": ""},
        })
        assert state.nodes["a1"].current_activity is None

    def test_tool_end_clears_current_activity(self):
        """tool_end resets current_activity to None."""
        state = self._make_state_with_running_agent()
        state.process_event({
            "event": "tool_start", "span_id": "s2",
            "data": {"agent_id": "a1", "tool": "Bash", "summary": "Running tests"},
        })
        assert state.nodes["a1"].current_activity == "Running tests"
        state.process_event({
            "event": "tool_end", "span_id": "s2",
            "data": {"agent_id": "a1", "tool": "Bash", "duration_ms": 500},
        })
        assert state.nodes["a1"].current_activity is None

    def test_tool_start_unknown_agent_ignored(self):
        """tool_start for a non-existent agent_id does not raise."""
        state = TreeState()
        state.process_event({
            "event": "tool_start", "span_id": "s2",
            "data": {"agent_id": "nonexistent", "tool": "Bash"},
        })
        assert "nonexistent" not in state.nodes

    def test_tool_end_unknown_agent_ignored(self):
        """tool_end for a non-existent agent_id does not raise."""
        state = TreeState()
        state.process_event({
            "event": "tool_end", "span_id": "s2",
            "data": {"agent_id": "nonexistent", "tool": "Bash"},
        })

    def test_tool_start_empty_agent_id_ignored(self):
        """tool_start with empty agent_id does nothing."""
        state = self._make_state_with_running_agent()
        state.process_event({
            "event": "tool_start", "span_id": "s2",
            "data": {"agent_id": "", "tool": "Bash", "summary": "x"},
        })
        assert state.nodes["a1"].current_activity is None

    def test_sequential_tool_starts_overwrite_activity(self):
        """A second tool_start overwrites the activity from the first."""
        state = self._make_state_with_running_agent()
        state.process_event({
            "event": "tool_start", "span_id": "s2",
            "data": {"agent_id": "a1", "tool": "Bash", "summary": "First"},
        })
        state.process_event({
            "event": "tool_start", "span_id": "s3",
            "data": {"agent_id": "a1", "tool": "Read", "summary": "Second"},
        })
        assert state.nodes["a1"].current_activity == "Second"

    def test_current_activity_included_in_to_dict(self):
        """current_activity field appears in serialized output."""
        state = self._make_state_with_running_agent()
        state.process_event({
            "event": "tool_start", "span_id": "s2",
            "data": {"agent_id": "a1", "tool": "Bash", "summary": "Building"},
        })
        result = state.to_dict()
        node = next(n for n in result["nodes"] if n["agent_id"] == "a1")
        assert node["current_activity"] == "Building"

    def test_current_activity_null_in_to_dict_by_default(self):
        """current_activity defaults to null in serialized output."""
        state = self._make_state_with_running_agent()
        result = state.to_dict()
        node = next(n for n in result["nodes"] if n["agent_id"] == "a1")
        assert node["current_activity"] is None

    def test_tool_start_on_completed_agent_still_sets_activity(self):
        """tool_start does not check node status — it sets activity even on completed agents.

        Note: this documents current behavior. set_activity() rejects
        non-running agents, but process_event(tool_start) does not.
        """
        state = self._make_state_with_running_agent()
        # Mark agent as completed
        state.process_event({
            "event": "agent_end", "span_id": "s1",
            "data": {"exit_code": 0, "duration_ms": 1000},
        })
        assert state.nodes["a1"].status == "completed"
        # tool_start still modifies it
        state.process_event({
            "event": "tool_start", "span_id": "s3",
            "data": {"agent_id": "a1", "tool": "Bash", "summary": "Late arrival"},
        })
        assert state.nodes["a1"].current_activity == "Late arrival"


class TestTreeStateSetActivity:
    """Tests for the set_activity() method used by log-based fallback."""

    def test_set_activity_on_running_agent(self):
        state = TreeState()
        state.load_session_json({
            "agents": {"a1": {"role": "impl", "runtime": "cc", "parent": None}},
        })
        assert state.set_activity("a1", "Reading file") is True
        assert state.nodes["a1"].current_activity == "Reading file"

    def test_set_activity_returns_false_when_unchanged(self):
        state = TreeState()
        state.load_session_json({
            "agents": {"a1": {"role": "impl", "runtime": "cc", "parent": None}},
        })
        state.set_activity("a1", "Reading file")
        assert state.set_activity("a1", "Reading file") is False

    def test_set_activity_returns_false_for_unknown_agent(self):
        state = TreeState()
        assert state.set_activity("nonexistent", "something") is False

    def test_set_activity_returns_false_for_non_running_agent(self):
        state = TreeState()
        state.load_session_json({
            "agents": {"a1": {"role": "impl", "runtime": "cc", "parent": None}},
        })
        state.nodes["a1"].status = "completed"
        assert state.set_activity("a1", "Reading file") is False

    def test_set_activity_clears_with_none(self):
        state = TreeState()
        state.load_session_json({
            "agents": {"a1": {"role": "impl", "runtime": "cc", "parent": None}},
        })
        state.set_activity("a1", "Reading file")
        assert state.set_activity("a1", None) is True
        assert state.nodes["a1"].current_activity is None

    def test_set_activity_clears_activity_action(self):
        """set_activity (log-fallback) clears activity_action since it has no type info."""
        state = TreeState()
        state.load_session_json({
            "agents": {"a1": {"role": "impl", "runtime": "cc", "parent": None}},
        })
        state.nodes["a1"].activity_action = "Read"
        state.set_activity("a1", "Doing something")
        assert state.nodes["a1"].activity_action is None


class TestComposeActivity:
    """Tests for the _compose_activity helper."""

    def test_action_only(self):
        assert _compose_activity("Think", "", "") == "Think"

    def test_action_and_target(self):
        assert _compose_activity("Read", "src/app.ts", "") == "Read src/app.ts"

    def test_action_target_detail(self):
        assert _compose_activity("Read", "src/app.ts", "lines 1-50") == "Read src/app.ts (lines 1-50)"

    def test_empty_action_returns_none(self):
        assert _compose_activity("", "target", "detail") is None

    def test_truncates_long_result(self):
        long_target = "x" * 200
        result = _compose_activity("Read", long_target, "")
        assert len(result) <= 121  # 120 + "…"
        assert result.endswith("…")


class TestTreeStateActivityUpdate:
    """Tests for activity_update event handling in TreeState."""

    def _make_state_with_running_agent(self, agent_id="a1"):
        """Helper: create a TreeState with one running agent."""
        state = TreeState()
        state.process_event({
            "event": "delegate_start", "span_id": "s1",
            "parent_span": None, "data": {"role": "impl"},
        })
        state.process_event({
            "event": "agent_spawn", "span_id": "s1", "ts": "T1",
            "data": {"agent_id": agent_id, "runtime": "cc", "pid": 1},
        })
        return state

    def test_activity_update_sets_current_activity(self):
        state = self._make_state_with_running_agent()
        state.process_event({
            "event": "activity_update", "span_id": "s2",
            "data": {"agent_id": "a1", "action": "Read", "target": "src/app.ts", "detail": ""},
        })
        assert state.nodes["a1"].current_activity == "Read src/app.ts"

    def test_activity_update_sets_activity_action(self):
        state = self._make_state_with_running_agent()
        state.process_event({
            "event": "activity_update", "span_id": "s2",
            "data": {"agent_id": "a1", "action": "Bash", "target": "npm test"},
        })
        assert state.nodes["a1"].activity_action == "Bash"

    def test_activity_update_with_detail(self):
        state = self._make_state_with_running_agent()
        state.process_event({
            "event": "activity_update", "span_id": "s2",
            "data": {"agent_id": "a1", "action": "Read", "target": "file.ts", "detail": "lines 1-50"},
        })
        assert state.nodes["a1"].current_activity == "Read file.ts (lines 1-50)"

    def test_activity_update_action_only(self):
        state = self._make_state_with_running_agent()
        state.process_event({
            "event": "activity_update", "span_id": "s2",
            "data": {"agent_id": "a1", "action": "Think"},
        })
        assert state.nodes["a1"].current_activity == "Think"
        assert state.nodes["a1"].activity_action == "Think"

    def test_activity_update_empty_action_sets_none(self):
        state = self._make_state_with_running_agent()
        state.process_event({
            "event": "activity_update", "span_id": "s2",
            "data": {"agent_id": "a1", "action": ""},
        })
        assert state.nodes["a1"].current_activity is None
        assert state.nodes["a1"].activity_action is None

    @patch("strawpot_gui.sse.time")
    def test_debounce_drops_rapid_events(self, mock_time):
        mock_time.monotonic.side_effect = [1.0, 1.3]  # 300ms apart
        state = self._make_state_with_running_agent()
        state.process_event({
            "event": "activity_update", "span_id": "s2",
            "data": {"agent_id": "a1", "action": "Read", "target": "file1.ts"},
        })
        state.process_event({
            "event": "activity_update", "span_id": "s3",
            "data": {"agent_id": "a1", "action": "Write", "target": "file2.ts"},
        })
        # Second event dropped — still shows first
        assert state.nodes["a1"].current_activity == "Read file1.ts"
        assert state.nodes["a1"].activity_action == "Read"

    @patch("strawpot_gui.sse.time")
    def test_debounce_accepts_after_threshold(self, mock_time):
        mock_time.monotonic.side_effect = [1.0, 1.6]  # 600ms apart
        state = self._make_state_with_running_agent()
        state.process_event({
            "event": "activity_update", "span_id": "s2",
            "data": {"agent_id": "a1", "action": "Read", "target": "file1.ts"},
        })
        state.process_event({
            "event": "activity_update", "span_id": "s3",
            "data": {"agent_id": "a1", "action": "Write", "target": "file2.ts"},
        })
        assert state.nodes["a1"].current_activity == "Write file2.ts"
        assert state.nodes["a1"].activity_action == "Write"

    def test_activity_update_unknown_agent_ignored(self):
        state = TreeState()
        state.process_event({
            "event": "activity_update", "span_id": "s2",
            "data": {"agent_id": "nonexistent", "action": "Read"},
        })
        assert "nonexistent" not in state.nodes

    def test_activity_update_empty_agent_ignored(self):
        state = self._make_state_with_running_agent()
        state.process_event({
            "event": "activity_update", "span_id": "s2",
            "data": {"agent_id": "", "action": "Read"},
        })
        assert state.nodes["a1"].current_activity is None

    def test_tool_end_clears_activity_action(self):
        state = self._make_state_with_running_agent()
        state.process_event({
            "event": "activity_update", "span_id": "s2",
            "data": {"agent_id": "a1", "action": "Read", "target": "file.ts"},
        })
        assert state.nodes["a1"].activity_action == "Read"
        state.process_event({
            "event": "tool_end", "span_id": "s3",
            "data": {"agent_id": "a1", "tool": "Read"},
        })
        assert state.nodes["a1"].activity_action is None
        assert state.nodes["a1"].current_activity is None

    def test_activity_action_in_to_dict(self):
        state = self._make_state_with_running_agent()
        state.process_event({
            "event": "activity_update", "span_id": "s2",
            "data": {"agent_id": "a1", "action": "Search", "target": "handleSubmit"},
        })
        result = state.to_dict()
        node = next(n for n in result["nodes"] if n["agent_id"] == "a1")
        assert node["activity_action"] == "Search"

    def test_activity_action_null_by_default(self):
        state = self._make_state_with_running_agent()
        result = state.to_dict()
        node = next(n for n in result["nodes"] if n["agent_id"] == "a1")
        assert node["activity_action"] is None

    def test_debounce_cleanup_on_agent_end(self):
        state = self._make_state_with_running_agent()
        state.process_event({
            "event": "activity_update", "span_id": "s2",
            "data": {"agent_id": "a1", "action": "Read", "target": "f.ts"},
        })
        assert "a1" in state._activity_debounce
        state.process_event({
            "event": "agent_end", "span_id": "s1",
            "data": {"exit_code": 0, "duration_ms": 1000},
        })
        assert "a1" not in state._activity_debounce

    def test_debounce_cleanup_on_delegate_end(self):
        state = self._make_state_with_running_agent()
        state.process_event({
            "event": "activity_update", "span_id": "s2",
            "data": {"agent_id": "a1", "action": "Read", "target": "f.ts"},
        })
        assert "a1" in state._activity_debounce
        state.process_event({
            "event": "delegate_end", "span_id": "s1",
            "data": {"exit_code": 0, "duration_ms": 1000},
        })
        assert "a1" not in state._activity_debounce



"""Project activity stats endpoint."""

from datetime import datetime, timedelta, timezone

from fastapi import APIRouter, Depends, HTTPException, Query

from strawpot_gui.db import get_db_conn

router = APIRouter(prefix="/api", tags=["stats"])

_VALID_PERIODS = {"7d": 7, "30d": 30, "90d": 90}


@router.get("/projects/{project_id}/stats")
def get_project_stats(
    project_id: int,
    period: str = Query("30d"),
    conn=Depends(get_db_conn),
):
    if period not in _VALID_PERIODS:
        raise HTTPException(422, f"Invalid period: {period}. Must be one of {', '.join(_VALID_PERIODS)}")

    # Verify project exists
    row = conn.execute("SELECT id FROM projects WHERE id = ?", (project_id,)).fetchone()
    if not row:
        raise HTTPException(404, "Project not found")

    days = _VALID_PERIODS[period]
    now = datetime.now(timezone.utc)
    since = (now - timedelta(days=days)).replace(hour=0, minute=0, second=0, microsecond=0)
    until = now.replace(hour=23, minute=59, second=59, microsecond=0)

    since_iso = since.isoformat()

    # Summary
    summary = conn.execute(
        """
        SELECT COUNT(*) AS total_runs,
               SUM(CASE WHEN status='completed' THEN 1 ELSE 0 END) AS completed,
               SUM(CASE WHEN status='failed' THEN 1 ELSE 0 END) AS failed,
               SUM(CASE WHEN status='stopped' THEN 1 ELSE 0 END) AS stopped,
               CAST(AVG(CASE WHEN duration_ms IS NOT NULL THEN duration_ms END) AS INTEGER) AS avg_duration_ms
        FROM sessions
        WHERE project_id = ? AND started_at >= ? AND status IN ('completed','failed','stopped')
        """,
        (project_id, since_iso),
    ).fetchone()

    total_runs = summary["total_runs"] or 0
    completed = summary["completed"] or 0
    failed = summary["failed"] or 0
    stopped = summary["stopped"] or 0
    avg_duration_ms = summary["avg_duration_ms"]
    # Success rate excludes stopped sessions — user-interrupted runs are
    # neither successes nor failures.
    decisive = completed + failed
    success_rate = round(completed / decisive * 100, 1) if decisive > 0 else 0.0

    # Daily breakdown
    daily_rows = conn.execute(
        """
        SELECT DATE(started_at) AS date, COUNT(*) AS total,
               SUM(CASE WHEN status='completed' THEN 1 ELSE 0 END) AS completed,
               SUM(CASE WHEN status='failed' THEN 1 ELSE 0 END) AS failed,
               SUM(CASE WHEN status='stopped' THEN 1 ELSE 0 END) AS stopped,
               CAST(AVG(CASE WHEN duration_ms IS NOT NULL THEN duration_ms END) AS INTEGER) AS avg_duration_ms
        FROM sessions
        WHERE project_id = ? AND started_at >= ? AND status IN ('completed','failed','stopped')
        GROUP BY DATE(started_at) ORDER BY date ASC
        """,
        (project_id, since_iso),
    ).fetchall()

    daily_map = {
        row["date"]: {
            "date": row["date"],
            "total": row["total"],
            "completed": row["completed"],
            "failed": row["failed"],
            "stopped": row["stopped"],
            "avg_duration_ms": row["avg_duration_ms"],
        }
        for row in daily_rows
    }

    # Gap-fill
    daily = []
    current = since.date()
    today = now.date()
    while current <= today:
        date_str = current.isoformat()
        daily.append(
            daily_map.get(date_str, {
                "date": date_str,
                "total": 0,
                "completed": 0,
                "failed": 0,
                "stopped": 0,
                "avg_duration_ms": None,
            })
        )
        current += timedelta(days=1)

    return {
        "period": period,
        "since": since.isoformat(),
        "until": until.isoformat(),
        "total_runs": total_runs,
        "completed": completed,
        "failed": failed,
        "stopped": stopped,
        "success_rate": success_rate,
        "avg_duration_ms": avg_duration_ms,
        "daily": daily,
    }



"""In-process pub/sub event bus for cross-session notifications."""

import asyncio
import json
from dataclasses import asdict, dataclass, field
from typing import AsyncIterator


@dataclass
class SessionEvent:
    """A global lifecycle event."""

    kind: str  # "session_started" | "session_completed" | "session_failed" | "session_stopped"
    run_id: str
    project_id: int | None = None
    data: dict = field(default_factory=dict)


class EventBus:
    """Simple broadcast pub/sub using asyncio.Queue per subscriber."""

    def __init__(self) -> None:
        self._subscribers: list[asyncio.Queue[SessionEvent]] = []

    def publish(self, event: SessionEvent) -> None:
        """Publish an event to all subscribers (non-blocking)."""
        for q in self._subscribers:
            try:
                q.put_nowait(event)
            except asyncio.QueueFull:
                pass  # drop if subscriber is too slow

    async def subscribe(self, poll_interval: float = 5.0) -> AsyncIterator[SessionEvent | None]:
        """Yield events as they arrive. Clean up on generator close.

        Yields ``None`` every *poll_interval* seconds when no event arrives,
        giving callers a chance to check for client disconnection.
        """
        q: asyncio.Queue[SessionEvent] = asyncio.Queue(maxsize=100)
        self._subscribers.append(q)
        try:
            while True:
                try:
                    event = await asyncio.wait_for(q.get(), timeout=poll_interval)
                    yield event
                except asyncio.TimeoutError:
                    yield None
        finally:
            self._subscribers.remove(q)


class ProgressEventAdapter:
    """Bridges CLI ProgressEvent to GUI EventBus.

    Maps ProgressEvent kinds to SessionEvent, preserving data
    for the GUI frontend to consume when ready.
    """

    def __init__(self, event_bus: EventBus, run_id: str, project_id: int | None = None) -> None:
        self._event_bus = event_bus
        self._run_id = run_id
        self._project_id = project_id

    def handle_event(self, event) -> None:
        """Callback for ``Session.on_event``."""
        self._event_bus.publish(SessionEvent(
            kind=f"progress_{event.kind}",
            run_id=self._run_id,
            project_id=self._project_id,
            data=asdict(event),
        ))


# Singleton instance, attached to app.state in lifespan
event_bus = EventBus()



"""Project-scoped resource endpoints — list, detail, install, uninstall."""

from pathlib import Path

from fastapi import APIRouter, Body, Depends, HTTPException

from strawpot.config import _read_toml, get_strawpot_home, save_resource_config

from strawpot_gui.db import get_db_conn
from strawpot_gui.routers.registry import (
    RESOURCE_TYPES,
    coerce_param,
    extract_saved_values,
    parse_manifest,
    read_version,
    run_strawhub,
    scan_dir,
    singular_type,
    validate_type,
)

router = APIRouter(prefix="/api/projects", tags=["project-resources"])


def _get_project_dir(project_id: int, conn) -> str:
    """Look up project working_dir from DB and validate it exists."""
    row = conn.execute(
        "SELECT working_dir FROM projects WHERE id = ?", (project_id,)
    ).fetchone()
    if not row:
        raise HTTPException(404, "Project not found")
    working_dir = row["working_dir"]
    if not Path(working_dir).is_dir():
        raise HTTPException(400, f"Project directory does not exist: {working_dir}")
    return working_dir


def _count_config_overrides(toml_data: dict, resource_type: str, name: str) -> int:
    """Count the number of env/param overrides for a resource in TOML data."""
    count = 0
    if resource_type == "roles":
        role_data = toml_data.get("roles", {}).get(name, {})
        count = len(role_data) if isinstance(role_data, dict) else 0
    elif resource_type == "skills":
        skill_data = toml_data.get("skills", {}).get(name, {})
        count = len(skill_data.get("env", {})) if isinstance(skill_data, dict) else 0
    elif resource_type == "agents":
        agent_data = toml_data.get("agents", {}).get(name, {})
        if not isinstance(agent_data, dict):
            agent_data = {}
        count = len(agent_data.get("env", {}))
        # Count non-env params (keys other than "env")
        count += sum(1 for k in agent_data if k != "env")
    elif resource_type == "memories":
        count = len(toml_data.get("memories", {}).get(name, {}).get("env", {}))
        # memory_config is shared across all memory providers, count if any
        count += len(toml_data.get("memory_config", {}))
    return count


@router.get("/{project_id}/resources")
def list_project_resources(project_id: int, conn=Depends(get_db_conn)):
    """List all resources available to a project (project-local + global).

    Project-local resources take precedence over global ones with the same name.
    Each resource includes a config_count of project-level overrides from
    strawpot.toml.
    """
    working_dir = _get_project_dir(project_id, conn)
    project_base = Path(working_dir) / ".strawpot"
    global_base = get_strawpot_home()
    toml_path = Path(working_dir) / "strawpot.toml"
    toml_data = _read_toml(toml_path)
    results = []
    for rtype, (dir_name, manifest) in RESOURCE_TYPES.items():
        # Scan project-local first, track names to avoid duplicates
        seen: set[str] = set()
        for item in scan_dir(project_base, dir_name, manifest, rtype, "project"):
            item["type"] = rtype
            item["config_count"] = _count_config_overrides(toml_data, rtype, item["name"])
            results.append(item)
            seen.add(item["name"])
        # Add global resources not shadowed by project-local ones
        for item in scan_dir(global_base, dir_name, manifest, rtype, "global"):
            if item["name"] not in seen:
                item["type"] = rtype
                item["config_count"] = _count_config_overrides(toml_data, rtype, item["name"])
                results.append(item)
    return results


def _resolve_resource_dir(
    working_dir: str, dir_name: str, manifest: str, name: str
) -> tuple[Path, str]:
    """Find a resource directory, checking project-local first then global.

    Returns (resource_dir, source) where source is "project" or "global".
    Raises HTTPException(404) if not found in either location.
    """
    project_dir = Path(working_dir) / ".strawpot" / dir_name / name
    if (project_dir / manifest).is_file():
        return project_dir, "project"
    global_dir = get_strawpot_home() / dir_name / name
    if (global_dir / manifest).is_file():
        return global_dir, "global"
    raise HTTPException(404, f"Resource not found: {dir_name}/{name}")


@router.get("/{project_id}/resources/{resource_type}/{name}")
def get_project_resource(
    project_id: int, resource_type: str, name: str, conn=Depends(get_db_conn)
):
    """Get detail for a resource available to this project."""
    working_dir = _get_project_dir(project_id, conn)
    dir_name, manifest = validate_type(resource_type)
    resource_dir, source = _resolve_resource_dir(working_dir, dir_name, manifest, name)

    fm, body = parse_manifest(resource_dir / manifest, resource_type)
    return {
        "name": fm.get("name", name),
        "version": read_version(resource_dir, fm),
        "description": fm.get("description", ""),
        "frontmatter": fm,
        "body": body,
        "source": source,
        "path": str(resource_dir),
    }


@router.get("/{project_id}/resources/{resource_type}/{name}/config")
def get_project_resource_config(
    project_id: int, resource_type: str, name: str, conn=Depends(get_db_conn)
):
    """Get env/params schema and saved values from project's strawpot.toml."""
    working_dir = _get_project_dir(project_id, conn)
    dir_name, manifest = validate_type(resource_type)
    resource_dir, _ = _resolve_resource_dir(working_dir, dir_name, manifest, name)

    fm, _ = parse_manifest(resource_dir / manifest, resource_type)
    strawpot_meta = fm.get("metadata", {}).get("strawpot", {})

    env_schema = strawpot_meta.get("env", {})
    params_schema = strawpot_meta.get("params", {})

    if resource_type == "roles":
        manifest_default = strawpot_meta.get("default_agent")
        params_schema = {
            "default_agent": {
                "type": "string",
                "default": manifest_default,
                "description": "Agent to use when running this role",
            },
        }

    toml_path = Path(working_dir) / "strawpot.toml"
    toml_data = _read_toml(toml_path)
    env_values, params_values = extract_saved_values(
        toml_data, resource_type, name, params_schema
    )

    return {
        "env_schema": env_schema,
        "env_values": env_values,
        "params_schema": params_schema,
        "params_values": params_values,
    }


@router.put("/{project_id}/resources/{resource_type}/{name}/config")
def put_project_resource_config(
    project_id: int,
    resource_type: str,
    name: str,
    data: dict = Body(...),
    conn=Depends(get_db_conn),
):
    """Save env and param values for a project-scoped resource."""
    working_dir = _get_project_dir(project_id, conn)
    dir_name, manifest = validate_type(resource_type)
    resource_dir, _ = _resolve_resource_dir(working_dir, dir_name, manifest, name)
    manifest_path = resource_dir / manifest

    env_values = data.get("env_values")
    params_values = data.get("params_values")

    if params_values:
        fm, _ = parse_manifest(manifest_path, resource_type)
        params_schema = fm.get("metadata", {}).get("strawpot", {}).get("params", {})
        coerced = {}
        for k, v in params_values.items():
            schema = params_schema.get(k, {})
            try:
                coerced[k] = coerce_param(v, schema.get("type"))
            except (ValueError, TypeError):
                coerced[k] = v
        params_values = coerced

    save_resource_config(
        Path(working_dir), resource_type, name, env_values, params_values
    )
    return {"ok": True}


@router.post("/{project_id}/resources/install")
def install_project_resource(
    project_id: int, data: dict = Body(...), conn=Depends(get_db_conn)
):
    """Install a resource to a project via strawhub --root."""
    working_dir = _get_project_dir(project_id, conn)
    resource_type = data.get("type", "")
    name = data.get("name", "")
    if not resource_type or not name:
        raise HTTPException(400, "Both 'type' and 'name' are required")
    singular = singular_type(resource_type)
    return run_strawhub("--root", working_dir, "install", singular, "-y", name)


@router.delete("/{project_id}/resources/{resource_type}/{name}")
def uninstall_project_resource(
    project_id: int, resource_type: str, name: str, conn=Depends(get_db_conn)
):
    """Uninstall a resource from a project via strawhub --root."""
    working_dir = _get_project_dir(project_id, conn)
    validate_type(resource_type)
    singular = singular_type(resource_type)
    return run_strawhub("--root", working_dir, "uninstall", singular, name)


@router.post("/{project_id}/resources/update")
def update_project_resource(
    project_id: int, data: dict = Body(...), conn=Depends(get_db_conn)
):
    """Update a project resource to its latest version via strawhub --root."""
    working_dir = _get_project_dir(project_id, conn)
    resource_type = data.get("type", "")
    name = data.get("name", "")
    if not resource_type or not name:
        raise HTTPException(400, "Both 'type' and 'name' are required")
    singular = singular_type(resource_type)
    return run_strawhub("--root", working_dir, "update", singular, "-y", name)


@router.post("/{project_id}/resources/update-all")
def update_all_project_resources(
    project_id: int, data: dict = Body(default={}), conn=Depends(get_db_conn)
):
    """Update all project resources to their latest versions via strawhub --root.

    Optionally accepts {"resource_type": "roles"} to scope the update to a
    single resource type.
    """
    working_dir = _get_project_dir(project_id, conn)
    args = ["--root", working_dir, "update", "--all", "-y"]
    resource_type = data.get("resource_type")
    if resource_type:
        validate_type(resource_type)
        args.extend(["--type", singular_type(resource_type)])
    return run_strawhub(*args, timeout=300)


@router.post("/{project_id}/resources/reinstall")
def reinstall_project_resource(
    project_id: int, data: dict = Body(...), conn=Depends(get_db_conn)
):
    """Reinstall a project resource (re-download current version)."""
    working_dir = _get_project_dir(project_id, conn)
    resource_type = data.get("type", "")
    name = data.get("name", "")
    if not resource_type or not name:
        raise HTTPException(400, "Both 'type' and 'name' are required")

    dir_name, _manifest = validate_type(resource_type)
    version_file = Path(working_dir) / ".strawpot" / dir_name / name / ".version"
    if not version_file.is_file():
        raise HTTPException(
            404, f"Resource not found or has no version: {resource_type}/{name}"
        )
    version = version_file.read_text(encoding="utf-8").strip()
    if not version:
        raise HTTPException(404, f"Empty version file for: {resource_type}/{name}")

    singular = singular_type(resource_type)
    return run_strawhub(
        "--root", working_dir, "install", singular, "-y", name,
        "--version", version, "--force",
    )



"""Tests for config read/write endpoints."""

import tomllib
from pathlib import Path


class TestGlobalConfig:
    def test_get_empty_global_config(self, client, tmp_path, monkeypatch):
        monkeypatch.setenv("STRAWPOT_HOME", str(tmp_path / "home"))
        resp = client.get("/api/config/global")
        assert resp.status_code == 200
        body = resp.json()
        assert body["values"] == {}
        assert "defaults" in body
        assert body["defaults"]["runtime"] == "strawpot-claude-code"

    def test_put_and_get_global_config(self, client, tmp_path, monkeypatch):
        monkeypatch.setenv("STRAWPOT_HOME", str(tmp_path / "home"))
        data = {"runtime": "codex"}

        resp = client.put("/api/config/global", json=data)
        assert resp.status_code == 200
        assert resp.json() == data

        # Verify file was written
        path = tmp_path / "home" / "strawpot.toml"
        assert path.is_file()
        with open(path, "rb") as f:
            written = tomllib.load(f)
        assert written == data

        # Verify round-trip via GET
        resp = client.get("/api/config/global")
        body = resp.json()
        assert body["values"] == data
        assert "defaults" in body


class TestProjectConfig:
    def _create_project(self, client, working_dir):
        resp = client.post("/api/projects", json={
            "display_name": "Test",
            "working_dir": str(working_dir),
        })
        return resp.json()["id"]

    def test_get_project_config_empty(self, client, tmp_path, monkeypatch):
        monkeypatch.setenv("STRAWPOT_HOME", str(tmp_path / "home"))
        project_dir = tmp_path / "proj"
        project_dir.mkdir()
        pid = self._create_project(client, project_dir)

        resp = client.get(f"/api/projects/{pid}/config")
        assert resp.status_code == 200
        data = resp.json()
        assert "merged" in data
        assert "project" in data
        assert "global" in data
        # No config files → project and global are empty dicts
        assert data["project"] == {}
        assert data["global"] == {}
        # Merged should have defaults
        assert data["merged"]["runtime"] == "strawpot-claude-code"

    def test_put_and_get_project_config(self, client, tmp_path, monkeypatch):
        monkeypatch.setenv("STRAWPOT_HOME", str(tmp_path / "home"))
        project_dir = tmp_path / "proj"
        project_dir.mkdir()
        pid = self._create_project(client, project_dir)

        config_data = {"runtime": "codex"}
        resp = client.put(f"/api/projects/{pid}/config", json=config_data)
        assert resp.status_code == 200

        # Verify file written
        path = project_dir / "strawpot.toml"
        assert path.is_file()

        # Verify merged config picks up the override
        resp = client.get(f"/api/projects/{pid}/config")
        data = resp.json()
        assert data["project"] == {"runtime": "codex"}
        assert data["merged"]["runtime"] == "codex"

    def test_merged_shows_both_sources(self, client, tmp_path, monkeypatch):
        home = tmp_path / "home"
        monkeypatch.setenv("STRAWPOT_HOME", str(home))

        # Write global config
        client.put("/api/config/global", json={"runtime": "global_rt"})

        # Create project with its own config
        project_dir = tmp_path / "proj"
        project_dir.mkdir()
        pid = self._create_project(client, project_dir)
        client.put(f"/api/projects/{pid}/config", json={"memory": "local"})

        resp = client.get(f"/api/projects/{pid}/config")
        data = resp.json()
        assert data["global"] == {"runtime": "global_rt"}
        assert data["project"] == {"memory": "local"}
        # Merged has global runtime + project memory
        assert data["merged"]["runtime"] == "global_rt"
        assert data["merged"]["memory"] == "local"

    def test_project_not_found(self, client):
        resp = client.get("/api/projects/999/config")
        assert resp.status_code == 404

    def test_put_project_not_found(self, client):
        resp = client.put("/api/projects/999/config", json={"runtime": "x"})
        assert resp.status_code == 404

    def test_put_preserves_skills_and_roles(self, client, tmp_path, monkeypatch):
        """Saving config form should not wipe skills/roles env values."""
        monkeypatch.setenv("STRAWPOT_HOME", str(tmp_path / "home"))
        project_dir = tmp_path / "proj"
        project_dir.mkdir()
        pid = self._create_project(client, project_dir)

        # Pre-populate toml with skills env values
        import tomli_w
        existing = {
            "runtime": "old-rt",
            "skills": {
                "notify-telegram": {"env": {"TOKEN": "secret"}},
                "browse": "*",
            },
            "roles": {"my-role": "*"},
        }
        with open(project_dir / "strawpot.toml", "wb") as f:
            tomli_w.dump(existing, f)

        # Save config form (only has runtime, no skills/roles)
        resp = client.put(f"/api/projects/{pid}/config", json={"runtime": "new-rt"})
        assert resp.status_code == 200

        # Verify skills and roles survived
        import tomllib
        with open(project_dir / "strawpot.toml", "rb") as f:
            written = tomllib.load(f)
        assert written["runtime"] == "new-rt"
        assert written["skills"]["notify-telegram"]["env"]["TOKEN"] == "secret"
        assert written["skills"]["browse"] == "*"
        assert written["roles"]["my-role"] == "*"

    def test_put_global_preserves_skills(self, client, tmp_path, monkeypatch):
        """Saving global config form should not wipe skills env values."""
        home = tmp_path / "home"
        monkeypatch.setenv("STRAWPOT_HOME", str(home))

        # Pre-populate global toml
        import tomli_w
        home.mkdir(parents=True, exist_ok=True)
        existing = {
            "runtime": "old",
            "agents": {"my-agent": {"env": {"KEY": "val"}}},
        }
        with open(home / "strawpot.toml", "wb") as f:
            tomli_w.dump(existing, f)

        # Save form data without agents section
        resp = client.put("/api/config/global", json={"runtime": "new"})
        assert resp.status_code == 200

        import tomllib
        with open(home / "strawpot.toml", "rb") as f:
            written = tomllib.load(f)
        assert written["runtime"] == "new"
        assert written["agents"]["my-agent"]["env"]["KEY"] == "val"



"""Tests for project file upload, listing, and deletion endpoints."""

import pytest


@pytest.fixture
def project(client, tmp_path):
    """Create a project with tmp_path as working_dir and return (id, tmp_path)."""
    resp = client.post(
        "/api/projects",
        json={"display_name": "files-proj", "working_dir": str(tmp_path)},
    )
    return resp.json()["id"], tmp_path


class TestListFiles:
    def test_empty_when_no_files_dir(self, client, project):
        pid, _ = project
        resp = client.get(f"/api/projects/{pid}/files")
        assert resp.status_code == 200
        assert resp.json() == []

    def test_lists_uploaded_files(self, client, project):
        pid, tmp_path = project
        files_dir = tmp_path / ".strawpot" / "files"
        files_dir.mkdir(parents=True)
        (files_dir / "a.txt").write_text("hello")
        (files_dir / "b.txt").write_text("world")

        resp = client.get(f"/api/projects/{pid}/files")
        assert resp.status_code == 200
        names = [f["name"] for f in resp.json()]
        assert "a.txt" in names
        assert "b.txt" in names

    def test_lists_nested_files(self, client, project):
        pid, tmp_path = project
        files_dir = tmp_path / ".strawpot" / "files"
        sub = files_dir / "sub"
        sub.mkdir(parents=True)
        (sub / "nested.txt").write_text("deep")

        resp = client.get(f"/api/projects/{pid}/files")
        entries = resp.json()
        assert len(entries) == 1
        assert entries[0]["name"] == "nested.txt"
        assert "sub" in entries[0]["path"]

    def test_nonexistent_project_returns_404(self, client):
        resp = client.get("/api/projects/999/files")
        assert resp.status_code == 404


class TestUploadFiles:
    def test_upload_single_file(self, client, project):
        pid, tmp_path = project
        resp = client.post(
            f"/api/projects/{pid}/files",
            files=[("files", ("test.txt", b"content", "text/plain"))],
        )
        assert resp.status_code == 201
        assert len(resp.json()) == 1
        assert resp.json()[0]["name"] == "test.txt"
        assert (tmp_path / ".strawpot" / "files" / "test.txt").is_file()

    def test_upload_multiple_files(self, client, project):
        pid, _ = project
        resp = client.post(
            f"/api/projects/{pid}/files",
            files=[
                ("files", ("a.txt", b"aaa", "text/plain")),
                ("files", ("b.txt", b"bbb", "text/plain")),
            ],
        )
        assert resp.status_code == 201
        assert len(resp.json()) == 2

    def test_path_traversal_rejected(self, client, project):
        pid, _ = project
        resp = client.post(
            f"/api/projects/{pid}/files",
            files=[("files", ("../escape.txt", b"bad", "text/plain"))],
        )
        assert resp.status_code == 400

    def test_absolute_path_rejected(self, client, project):
        pid, _ = project
        resp = client.post(
            f"/api/projects/{pid}/files",
            files=[("files", ("/etc/passwd", b"bad", "text/plain"))],
        )
        assert resp.status_code == 400

    def test_nonexistent_project_returns_404(self, client):
        resp = client.post(
            "/api/projects/999/files",
            files=[("files", ("test.txt", b"x", "text/plain"))],
        )
        assert resp.status_code == 404


class TestDeleteFile:
    def test_delete_file(self, client, project):
        pid, tmp_path = project
        files_dir = tmp_path / ".strawpot" / "files"
        files_dir.mkdir(parents=True)
        (files_dir / "doomed.txt").write_text("bye")

        resp = client.delete(f"/api/projects/{pid}/files/doomed.txt")
        assert resp.status_code == 200
        assert resp.json()["ok"] is True
        assert not (files_dir / "doomed.txt").exists()

    def test_delete_cleans_empty_parents(self, client, project):
        pid, tmp_path = project
        files_dir = tmp_path / ".strawpot" / "files"
        sub = files_dir / "sub"
        sub.mkdir(parents=True)
        (sub / "file.txt").write_text("x")

        client.delete(f"/api/projects/{pid}/files/sub/file.txt")
        assert not sub.exists()

    def test_file_not_found_returns_404(self, client, project):
        pid, tmp_path = project
        files_dir = tmp_path / ".strawpot" / "files"
        files_dir.mkdir(parents=True)

        resp = client.delete(f"/api/projects/{pid}/files/nope.txt")
        assert resp.status_code == 404

    def test_path_traversal_rejected(self, client, project):
        pid, tmp_path = project
        files_dir = tmp_path / ".strawpot" / "files"
        files_dir.mkdir(parents=True)
        # Create a file outside files_dir to attempt to reach
        (tmp_path / ".strawpot" / "secret.txt").write_text("secret")
        resp = client.delete(f"/api/projects/{pid}/files/sub/../../../secret.txt")
        # FastAPI or the handler rejects traversal (400, 404, or 422)
        assert resp.status_code >= 400

    def test_nonexistent_project_returns_404(self, client):
        resp = client.delete("/api/projects/999/files/test.txt")
        assert resp.status_code == 404



"""Tests for POST /api/sessions (session launch) endpoint."""

from unittest.mock import patch

from test_sessions_sync import _register_project

from strawpot_gui.db import get_db


class TestLaunchSession:
    def _launch(self, client, pid, **json_overrides):
        """POST /api/sessions with mocked subprocess and shutil.which."""
        body = {"project_id": pid, "task": "Fix the tests"}
        body.update(json_overrides)
        with patch("strawpot_gui.routers.sessions.shutil.which", return_value="/usr/bin/strawpot"), \
             patch("strawpot_gui.routers.sessions.subprocess.Popen") as mock_popen, \
             patch("strawpot_gui.routers.sessions.load_config") as mock_config:
            from strawpot.config import StrawPotConfig
            mock_config.return_value = StrawPotConfig()
            resp = client.post("/api/sessions", json=body)
        return resp, mock_popen

    def test_launch_returns_201(self, client, tmp_path):
        """Successful launch returns 201 with run_id and status."""
        project_dir = tmp_path / "proj"
        project_dir.mkdir()
        pid = _register_project(client, project_dir)

        resp, mock_popen = self._launch(client, pid)

        assert resp.status_code == 201
        data = resp.json()
        assert data["run_id"].startswith("run_")
        assert data["status"] == "starting"
        mock_popen.assert_called_once()

    def test_launch_passes_run_id_to_cli(self, client, tmp_path):
        """The subprocess command includes --run-id with the pre-generated ID."""
        project_dir = tmp_path / "proj"
        project_dir.mkdir()
        pid = _register_project(client, project_dir)

        resp, mock_popen = self._launch(client, pid)

        run_id = resp.json()["run_id"]
        cmd = mock_popen.call_args[0][0]
        assert "--run-id" in cmd
        assert run_id in cmd
        assert "--headless" in cmd
        assert "--task" in cmd

    def test_launch_with_overrides(self, client, tmp_path):
        """Overrides are passed as CLI flags."""
        project_dir = tmp_path / "proj"
        project_dir.mkdir()
        pid = _register_project(client, project_dir)

        resp, mock_popen = self._launch(
            client, pid,
            role="deployer",
            overrides={"runtime": "gemini"},
        )

        assert resp.status_code == 201
        cmd = mock_popen.call_args[0][0]
        assert "--role" in cmd
        assert "deployer" in cmd
        assert "--runtime" in cmd
        assert "gemini" in cmd

    def test_launch_inserts_db_row(self, client, tmp_path):
        """A DB row with status 'starting' is inserted."""
        project_dir = tmp_path / "proj"
        project_dir.mkdir()
        pid = _register_project(client, project_dir)

        resp, _ = self._launch(client, pid, task="Build feature")

        run_id = resp.json()["run_id"]
        with get_db(client.app.state.db_path) as conn:
            row = conn.execute(
                "SELECT * FROM sessions WHERE run_id = ?", (run_id,)
            ).fetchone()
        assert row is not None
        assert row["status"] == "starting"
        assert row["task"] == "Build feature"
        assert row["project_id"] == pid

    def test_launch_nonexistent_project_returns_404(self, client):
        """Unknown project_id returns 404."""
        resp = client.post("/api/sessions", json={
            "project_id": 9999,
            "task": "anything",
        })
        assert resp.status_code == 404

    def test_launch_missing_dir_returns_422(self, client, tmp_path):
        """Project whose working_dir no longer exists returns 422."""
        resp = client.post("/api/projects", json={
            "display_name": "Ghost",
            "working_dir": str(tmp_path / "vanished"),
        })
        pid = resp.json()["id"]

        with patch("strawpot_gui.routers.sessions.load_config"):
            resp = client.post("/api/sessions", json={
                "project_id": pid,
                "task": "doomed",
            })
        assert resp.status_code == 422

    def test_launch_empty_task_returns_422(self, client, tmp_path):
        """Empty or whitespace-only task is rejected."""
        project_dir = tmp_path / "proj"
        project_dir.mkdir()
        pid = _register_project(client, project_dir)

        resp = client.post("/api/sessions", json={
            "project_id": pid,
            "task": "   ",
        })
        assert resp.status_code == 422

    def test_launch_strawpot_not_on_path_returns_500(self, client, tmp_path):
        """If strawpot binary is not on PATH, returns 500."""
        project_dir = tmp_path / "proj"
        project_dir.mkdir()
        pid = _register_project(client, project_dir)

        with patch("strawpot_gui.routers.sessions.shutil.which", return_value=None), \
             patch("strawpot_gui.routers.sessions.load_config") as mock_config:
            from strawpot.config import StrawPotConfig
            mock_config.return_value = StrawPotConfig()

            resp = client.post("/api/sessions", json={
                "project_id": pid,
                "task": "anything",
            })
        assert resp.status_code == 500
